import { Any, O } from 'ts-toolbelt';

declare class SubmoduleController {
}
type ControllerSignal = SubmoduleController | Promise<SubmoduleController>;
declare class StopWithError extends SubmoduleController {
    phase: 'definition' | 'config' | 'services' | 'route' | 'router' | 'start' | 'runtime';
    exitCode: number;
    exitReason: string;
    error: any;
    constructor(phase: 'definition' | 'config' | 'services' | 'route' | 'router' | 'start' | 'runtime', exitCode: number, exitReason: string, error: any);
}
declare const stopAndDisplayHelp: (helpMessage: string, exitCode?: number) => SubmoduleController;
declare const stopWithError: (phase: StopWithError['phase'], exitCode?: StopWithError['exitCode'], exitReason?: StopWithError['exitReason'], error?: any) => SubmoduleController;

interface IOExtractor<T = unknown> {
    routeModule: T;
    input: unknown;
    output: unknown;
}
type Submodule<InitArgs = {}, Config = {}, Services = {}> = Any.Compute<{
    createConfig?: (createConfigParams?: {
        initArgs: InitArgs;
    }) => Config | Promise<Config>;
    createServices?: (createServiceParams: {
        initArgs: InitArgs;
        config: Config;
    }) => Services | Promise<Services>;
}, "flat">;
type ExecutableSubmodule<InitArgs = {}, Config = {}, Services = {}, RouteModule = unknown, Input = unknown, Output = unknown> = Submodule<InitArgs, Config, Services> & Any.Compute<{
    loadRouteModule?: (loadRouteModuleParams: {
        initArgs: InitArgs;
        config: Config;
        query: string;
        preloadedRoutes?: any;
    }) => RouteModule | Promise<RouteModule> | undefined | Promise<undefined>;
    execute: (executeParams: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        route: RouteModule;
        query?: string;
        input?: Input;
    }) => Promise<Output>;
}, "flat">;
type RouteLike<Context = unknown, RouteModule = unknown> = {
    handle: (context: Context) => unknown | Promise<unknown>;
    routeModule: RouteModule;
    routeName: string;
};
type ServerSubmodule<InitArgs = {}, Config = {}, Services = {}, Context = unknown, RouteModule = unknown, Route extends RouteLike<Context, RouteModule> = RouteLike<Context, RouteModule>, Router extends Record<string, Route> = Record<string, Route>> = Submodule<InitArgs, Config, Services> & Any.Compute<{
    loadRouteModules?: (loadRouteModuleParams: {
        initArgs: InitArgs;
        config: Config;
    }) => Record<string, RouteModule> | Promise<Record<string, RouteModule>>;
    createRoute: <OptionalRoute extends O.Optional<Route, "routeName" | "routeModule">>(createRouteParams: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        routeModule: RouteModule;
        routeName: string;
    }) => Promise<OptionalRoute | OptionalRoute[] | undefined> | (OptionalRoute | OptionalRoute[] | undefined);
    createRouter?: (createRouterParams: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        routes: Record<string, Route>;
    }) => Router | Promise<Router>;
    createInnerClient?: (createInnerClientParams: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        router: Router;
        startPromise: Promise<any>;
    }) => Promise<(query: string, input: unknown) => unknown | Promise<unknown>>;
    serve: (input: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        router: Router;
    }) => Promise<void>;
}, "flat">;
type GetRouteModules<S extends ServerSubmodule<any, any, any, any, any, any>> = S["loadRouteModules"] extends (...args: any[]) => infer T ? Awaited<T> : never;
declare const createExecutable: <Extractor extends IOExtractor<unknown>, Routes = unknown>(submoduleDef: ExecutableSubmodule<any, any, any, any>, initArgs: any) => <Query extends keyof Routes, IO extends Extractor & {
    routeModule: Routes[Query];
} = Extractor & {
    routeModule: Routes[Query];
}>(query: Query | Omit<string, Query>, input: IO["input"]) => Promise<IO["output"]>;
interface ExtractRouteFn<RouteModule = unknown> {
    routeModule: RouteModule;
    routeFn: unknown;
}
declare const serve: (nonValidatedSubmodule: ServerSubmodule<any, any, any, any, any, any, any>, initArgs: any) => Promise<{
    submodule: {
        createConfig: (createConfigParams?: {
            initArgs: any;
        } | undefined) => any;
        createServices: (createServiceParams: {
            initArgs: any;
            config: any;
        }) => any;
        loadRouteModules: (loadRouteModuleParams: {
            initArgs: any;
            config: any;
        }) => Record<string, any> | Promise<Record<string, any>>;
        createRoute: <OptionalRoute extends {
            routeName?: any;
            routeModule?: any;
        } | {
            [x: string]: unknown;
            [x: number]: any;
            [x: symbol]: unknown;
            routeName?: any;
            routeModule?: any;
        }>(createRouteParams: {
            initArgs: any;
            config: any;
            services: any;
            routeModule: any;
            routeName: string;
        }) => OptionalRoute | OptionalRoute[] | Promise<OptionalRoute | OptionalRoute[] | undefined> | undefined;
        createRouter: (createRouterParams: {
            initArgs: any;
            config: any;
            services: any;
            routes: Record<string, any>;
        }) => any;
        createInnerClient: (createInnerClientParams: {
            initArgs: any;
            config: any;
            services: any;
            router: any;
            startPromise: Promise<any>;
        }) => Promise<(query: string, input: unknown) => unknown | Promise<unknown>>;
        serve: (input: {
            initArgs: any;
            config: any;
            services: any;
            router: any;
        }) => Promise<void>;
        loadRouteModule(): Promise<undefined>;
        execute(): Promise<never>;
    };
    config: any;
    services: any;
    router: any;
    startPromise: Promise<void>;
    client: (query: string, input: unknown) => unknown | Promise<unknown>;
}>;
type DefaultRouteModule<CallContext = {}, Output = unknown> = {
    default: (callContext: CallContext) => Output | Promise<Output>;
};
interface DefaultRouteFnExtractor extends ExtractRouteFn<DefaultRouteModule> {
    routeFn: this["routeModule"];
}
interface DefaultInputOutputExtractor<RouteModule extends DefaultRouteModule> extends IOExtractor<RouteModule> {
    input: Parameters<this["routeModule"]["default"]>[0] extends {
        input: infer Input;
    } ? Input : unknown;
    output: ReturnType<this["routeModule"]["default"]> extends Promise<infer Result> ? Result : ReturnType<this["routeModule"]["default"]>;
}
declare class SubmoduleBuilder<InitArgs = {}, Config = {}, Services = {}, Context = {}, RouteModule = unknown, Route extends RouteLike<Context, RouteModule> = RouteLike<Context, RouteModule>, RouteFn = unknown, Routes = unknown, InputOutputExtractor extends IOExtractor = IOExtractor> {
    defaults(): SubmoduleBuilder<InitArgs, Config, Services, Context, DefaultRouteModule<{
        config: Config;
        services: Services;
        context: Context;
    }>, RouteLike<Context, DefaultRouteModule<{
        config: Config;
        services: Services;
        context: Context;
    }>>, (DefaultRouteFnExtractor & {
        routeModule: DefaultRouteModule<{
            config: Config;
            services: Services;
            context: Context;
        }>;
    })["routeFn"], Routes, DefaultInputOutputExtractor<DefaultRouteModule<{
        config: Config;
        services: Services;
        context: Context;
    }>>>;
    routes<Routes>(): SubmoduleBuilder<InitArgs, Config, Services, Context, RouteModule, Route, RouteFn, Routes, InputOutputExtractor>;
    init<InitType>(): SubmoduleBuilder<InitType, Config, Services, Context, RouteModule, Route, RouteFn, Routes, InputOutputExtractor>;
    config<ConfigType>(): SubmoduleBuilder<InitArgs, ConfigType, Services, Context, RouteModule, Route, RouteFn, Routes, InputOutputExtractor>;
    services<ServicesType>(): SubmoduleBuilder<InitArgs, Config, ServicesType, Context, RouteModule, Route, RouteFn, Routes, InputOutputExtractor>;
    context<ContextType>(): SubmoduleBuilder<InitArgs, Config, Services, ContextType, RouteModule, RouteLike<ContextType, RouteModule>, RouteFn, Routes, InputOutputExtractor>;
    routeModule<RouteModuleType, Extractor extends ExtractRouteFn, ExtractInputOut extends IOExtractor<RouteModuleType>>(): SubmoduleBuilder<InitArgs, Config, Services, Context, RouteModuleType, RouteLike<Context, RouteModuleType>, (Extractor & {
        routeModule: RouteModuleType;
    })["routeFn"], Routes, ExtractInputOut>;
    route<RouteType extends RouteLike<Context, RouteModule>>(): SubmoduleBuilder<InitArgs, Config, Services, Context, RouteModule, RouteType, RouteFn, Routes, InputOutputExtractor>;
    serverSubmodule: ServerSubmodule<InitArgs, Config, Services, Context, RouteModule, Route, Record<string, Route>>;
    executableSubmodule: ExecutableSubmodule<InitArgs, Config, Services, RouteModule>;
    types: {
        initArgs: InitArgs;
        config: Config;
        services: Services;
        context: Context;
        route: Route;
        routeFn: RouteFn;
        routeModule: RouteModule;
        queries: keyof Routes;
    };
    defineRouteFn<RouteFnType extends RouteFn>(routeFn: RouteFnType): RouteFnType;
    defineExecutable(submoduleDef: ExecutableSubmodule<InitArgs, Config, Services, RouteModule>): ExecutableSubmodule<InitArgs, Config, Services, RouteModule, unknown, unknown>;
    createExecutable<Extractor extends IOExtractor>(submoduleDef: ExecutableSubmodule<InitArgs, Config, Services, RouteModule>, initArgs?: InitArgs): <Query extends keyof Routes, IO extends Extractor & {
        routeModule: Routes[Query];
    } = Extractor & {
        routeModule: Routes[Query];
    }>(query: Query | Omit<string, Query>, input: IO["input"]) => Promise<IO["output"]>;
    defineServer(submoduleDef: ServerSubmodule<InitArgs, Config, Services, Context, RouteModule>): ServerSubmodule<InitArgs, Config, Services, Context, RouteModule, RouteLike<Context, RouteModule>, Record<string, RouteLike<Context, RouteModule>>>;
    serve(submoduleDef: ServerSubmodule<InitArgs, Config, Services, Context, RouteModule>, initArgs?: InitArgs): Promise<{
        submodule: {
            createConfig: (createConfigParams?: {
                initArgs: any;
            } | undefined) => any;
            createServices: (createServiceParams: {
                initArgs: any;
                config: any;
            }) => any;
            loadRouteModules: (loadRouteModuleParams: {
                initArgs: any;
                config: any;
            }) => Record<string, any> | Promise<Record<string, any>>;
            createRoute: <OptionalRoute extends {
                routeName?: any;
                routeModule?: any;
            } | {
                [x: string]: unknown;
                [x: number]: any;
                [x: symbol]: unknown;
                routeName?: any;
                routeModule?: any;
            }>(createRouteParams: {
                initArgs: any;
                config: any;
                services: any;
                routeModule: any;
                routeName: string;
            }) => OptionalRoute | OptionalRoute[] | Promise<OptionalRoute | OptionalRoute[] | undefined> | undefined;
            createRouter: (createRouterParams: {
                initArgs: any;
                config: any;
                services: any;
                routes: Record<string, any>;
            }) => any;
            createInnerClient: (createInnerClientParams: {
                initArgs: any;
                config: any;
                services: any;
                router: any;
                startPromise: Promise<any>;
            }) => Promise<(query: string, input: unknown) => unknown>;
            serve: (input: {
                initArgs: any;
                config: any;
                services: any;
                router: any;
            }) => Promise<void>;
            loadRouteModule(): Promise<undefined>;
            execute(): Promise<never>;
        };
        config: any;
        services: any;
        router: any;
        startPromise: Promise<void>;
        client: (query: string, input: unknown) => unknown;
    }>;
}
declare const builder: () => SubmoduleBuilder<{}, {}, {}, {}, unknown, RouteLike<{}, unknown>, unknown, unknown, IOExtractor<unknown>>;

export { ControllerSignal, DefaultInputOutputExtractor, DefaultRouteFnExtractor, DefaultRouteModule, ExecutableSubmodule, ExtractRouteFn, GetRouteModules, IOExtractor, RouteLike, ServerSubmodule, Submodule, SubmoduleController, builder, createExecutable, serve, stopAndDisplayHelp, stopWithError };
