var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  // If the importer is in node compatibility mode or this is not an ESM
  // file that has been converted to a CommonJS file using a Babel-
  // compatible transform (i.e. "__esModule" has not been set), then set
  // "default" to the CommonJS "module.exports" for node compatibility.
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  SubmoduleController: () => SubmoduleController,
  builder: () => builder,
  createExecutable: () => createExecutable,
  serve: () => serve,
  stopAndDisplayHelp: () => stopAndDisplayHelp,
  stopWithError: () => stopWithError
});
module.exports = __toCommonJS(src_exports);

// src/controller.ts
var import_util = __toESM(require("util"));
var SubmoduleController = class {
};
var DisplayHelp = class extends SubmoduleController {
  constructor(helpMessage, exitCode) {
    super();
    this.helpMessage = helpMessage;
    this.exitCode = exitCode;
  }
};
var StopWithError = class extends SubmoduleController {
  constructor(phase, exitCode, exitReason, error) {
    super();
    this.phase = phase;
    this.exitCode = exitCode;
    this.exitReason = exitReason;
    this.error = error;
  }
};
var stopAndDisplayHelp = (helpMessage, exitCode = 0) => {
  return new DisplayHelp(helpMessage, exitCode);
};
var stopWithError = (phase, exitCode, exitReason, error) => {
  return new StopWithError(
    phase,
    exitCode || 1,
    exitReason || "",
    error
  );
};
function withControllerUnit(fn) {
  return function() {
    const that = this;
    const args = arguments;
    const result = fn.apply(that, args);
    if (result === void 0) {
      return;
    } else if (import_util.default.types.isPromise(result)) {
      return result.then((next) => {
        return next;
      }).catch((e) => {
        if (e instanceof DisplayHelp) {
          console.log(e.helpMessage);
          process.exit(e.exitCode);
        }
        if (!(e instanceof StopWithError)) {
          e = new StopWithError("runtime", 1, "Unexpected exception", e);
        }
        console.log("Caught an exception %s", e);
        process.exit(e.exitCode);
      });
    } else {
      return result;
    }
  };
}

// src/core.ts
var import_debug2 = __toESM(require("debug"));

// src/instrument.ts
var import_api = __toESM(require("@opentelemetry/api"));
var import_semantic_conventions = require("@opentelemetry/semantic-conventions");
var import_util2 = __toESM(require("util"));
var import_debug = __toESM(require("debug"));
function getTracer() {
  return import_api.default.trace.getTracer("submodule");
}
function trace(name, fn, spanOptions = {}) {
  return function() {
    const that = this;
    const args = arguments;
    return getTracer().startActiveSpan(name, spanOptions, (span) => {
      try {
        span.setAttribute(import_semantic_conventions.SemanticAttributes.CODE_FUNCTION, name);
        const result = fn.apply(that, args);
        if (import_util2.default.types.isPromise(result)) {
          return result.then((next) => {
            span.setStatus({ code: import_api.SpanStatusCode.OK });
            return next;
          });
        } else {
          span.setStatus({ code: import_api.SpanStatusCode.OK });
          return result;
        }
      } catch (e) {
        span.setStatus({ code: import_api.SpanStatusCode.ERROR }).recordException(e);
      } finally {
        span.end();
      }
    });
  };
}
function debug(name, fn) {
  const _debugger = (0, import_debug.default)(`submodule.runtime.${name}`);
  return function() {
    const that = this;
    const args = arguments;
    _debugger("calling function %s, with %O", name, arguments);
    const result = fn.apply(that, args);
    if (result === void 0) {
      _debugger("completed function call %s", name);
      return;
    } else if (import_util2.default.types.isPromise(result)) {
      return result.then((next) => {
        _debugger("completed function call %s, with %O", name, next);
        return next;
      }).catch((e) => {
        _debugger("caught an exception function call %s, with %O", name, e);
        throw e;
      });
    } else {
      return result;
    }
  };
}
function instrument(input, instrumentFn, maxLevel = 1, currentLevel = 0) {
  if (input === null || input === void 0)
    return input;
  if (currentLevel > maxLevel)
    return input;
  if (
    // there can be more
    typeof input !== "object" || Array.isArray(input) || // built-in objects like date, bytearray etc
    import_util2.default.types.isDate(input)
  ) {
    return input;
  }
  Object.keys(input).forEach((key) => {
    const instance = input[key];
    if (typeof instance === "function") {
      input[key] = instrumentFn(key, instance);
    } else {
      input[key] = instrument(input[key], instrumentFn, maxLevel, currentLevel + 1);
    }
  });
  return input;
}

// src/core.ts
var createExecutable = function(submoduleDef, initArgs) {
  let cached = void 0;
  const routeModules = /* @__PURE__ */ new Map();
  const submodule = { ...defaultSubmodule, ...submoduleDef };
  instrument(submodule, debug);
  instrument(submodule, trace);
  async function load() {
    if (!cached) {
      const config = await submodule.createConfig({ initArgs });
      const services = await submodule.createServices({ config, initArgs });
      cached = { config, services };
    }
    return cached;
  }
  return withControllerUnit(async (query, input) => {
    const queryString = query.toString();
    const { config, services } = await load();
    const routeModule = routeModules.has(queryString) ? routeModules.get(query) : await submodule.loadRouteModule({
      initArgs,
      config,
      query: queryString
    });
    return await submodule.execute({
      initArgs,
      config,
      services,
      query: queryString,
      input,
      route: routeModule
    });
  });
};
var serve = withControllerUnit(
  async (nonValidatedSubmodule, initArgs) => {
    const _debug = (0, import_debug2.default)("submodule.createSubmoduleInstance");
    const submodule = { ...defaultSubmodule, ...nonValidatedSubmodule };
    instrument(submodule, debug);
    instrument(submodule, trace);
    instrument(submodule, (name, fn) => withControllerUnit(fn));
    const config = await submodule.createConfig({ initArgs });
    const routeModules = await submodule.loadRouteModules({ config, initArgs });
    const services = await submodule.createServices({ config, initArgs });
    const routes = {};
    for (const [routeName, routeModule] of Object.entries(routeModules)) {
      const maybeRoute = await submodule.createRoute({
        initArgs,
        config,
        services,
        routeModule,
        routeName
      });
      instrument(maybeRoute, debug, 2);
      instrument(maybeRoute, trace, 2);
      if (maybeRoute === void 0)
        continue;
      if (Array.isArray(maybeRoute)) {
        maybeRoute.forEach((route) => routes[route["routeName"]] = route);
      } else {
        routes[maybeRoute["routeName"]] = maybeRoute;
      }
    }
    const router = await submodule.createRouter({
      initArgs,
      config,
      services,
      routes
    });
    const startPromise = submodule.serve({
      initArgs,
      config,
      services,
      router
    });
    const client = await submodule.createInnerClient({
      initArgs,
      config,
      services,
      router,
      startPromise
    });
    return { submodule, config, services, router, startPromise, client };
  }
);
var SubmoduleBuilder = class {
  defaults() {
    return this;
  }
  routes() {
    return this;
  }
  init() {
    return this;
  }
  config() {
    return this;
  }
  services() {
    return this;
  }
  context() {
    return this;
  }
  routeModule() {
    return this;
  }
  route() {
    return this;
  }
  defineRouteFn(routeFn) {
    return routeFn;
  }
  // #region executable
  defineExecutable(submoduleDef) {
    return submoduleDef;
  }
  createExecutable(submoduleDef, initArgs) {
    return createExecutable(submoduleDef, initArgs);
  }
  // #endregion
  // #region server
  defineServer(submoduleDef) {
    return submoduleDef;
  }
  serve(submoduleDef, initArgs) {
    return serve(submoduleDef, initArgs);
  }
  // #endregion
};
var builder = () => new SubmoduleBuilder();
var defaultSubmoduleDef = builder().defaults();
var defaultSubmodule = {
  async createConfig() {
    return {};
  },
  async createServices() {
    return {};
  },
  async loadRouteModule() {
    return void 0;
  },
  async loadRouteModules() {
    return {};
  },
  async createRoute() {
    throw new Error("not yet implemented");
  },
  async createRouter({ routes }) {
    return routes;
  },
  async execute() {
    throw new Error("not yet implemented");
  },
  async serve() {
    throw new Error("not yet implemented");
  },
  async createInnerClient({ startPromise }) {
    await startPromise;
    return () => {
      throw new Error("not yet implemented");
    };
  }
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  SubmoduleController,
  builder,
  createExecutable,
  serve,
  stopAndDisplayHelp,
  stopWithError
});
//# sourceMappingURL=index.js.map